import InitialSchedules.*;
import ScheduleUtil.*;
import SchedulesRecovery.*;
import ModifiedSchedules.*;
import jdistlib.rng.MersenneTwister;
import jdistlib.rng.RandomEngine;
import org.chocosolver.solver.Solution;

import java.util.ArrayList;
import java.util.List;
import java.util.PriorityQueue;

@SuppressWarnings("unused")
public class MainClass {

    public static void main(String[] args) {
        
        ParametersSchedules myScheduleParameters = new InitialSchedules.ParametersSchedules();
        for (int i = 0 ; i <= 30 ; i++) {
            
//            ParametersAbsencesSimulator.probPresence = 1 - (double) i * 0.01; // Pour de 0.90 a 1.0 par increment de 0.01
            
//            for (int j = 0 ; j <= 10 ; j++) {
                
//                ParametersVariableDemandSimulator.probIncreasedDemandbyOne = (double) j * 0.1; // Pour de 0 a 1 par increment de 0.1
//                ParametersVariableDemandSimulator.probIncreasedDemandbyTwo = ParametersVariableDemandSimulator.probIncreasedDemandbyOne/5; // Pour de 0 a 0.25 par increment de 0.05
//                ParametersVariableDemandSimulator.probIncreasedDemandbyThree = ParametersVariableDemandSimulator.probIncreasedDemandbyOne/25; // Pour de 0 a 0.25 par increment de 0.05
                List<SchedulesArray> initialSchedulesArrayList = generateInitialSchedules(myScheduleParameters);

                List<BankModifiedSchedules> listBankModifiedSchedules = GenerateModifiedSchedules(initialSchedulesArrayList, myScheduleParameters);

                List<BankRecoveredSchedules> bankRecoveredSchedulesOrderedByMeanCost = GenerateOptimalRecoveredSchedules(listBankModifiedSchedules);

//                SchedulesWindows.ShowSolutionResultsFrame initialSchedulesViewer = new SchedulesWindows.ShowSolutionResultsFrame(bankRecoveredSchedulesOrderedByMeanCost);
            
//            int optimalFullTimeEmployee = bankRecoveredSchedulesOrderedByMeanCost.get(0).initialSchedulesArray.getNumberFullTimeEmployee();
//            System.out.print(optimalFullTimeEmployee + " ");
//            }
//            System.out.println(" ");
        }
    }
    
    @SuppressWarnings("unused")
    private static List<SchedulesArray> generateInitialSchedules(ParametersSchedules myScheduleParameters) {

        // Trouver les horaires des employes a temps plein possibles pour different ratio templein/temps partiel
        List<SchedulesArray> initialSchedulesArrayList = new ArrayList<>();
        
        int minFullTimeEmployee = 9; //La solution avec 1 employes est trop lente
        
        int maxFullTimeEmployee = 9; //9 est le max d'employes a temps plein pour la presente requiredWorkforce
        
        int simulationNumber = 1;
        
        for (int numberFullTimeEmployees = minFullTimeEmployee ; numberFullTimeEmployees <= maxFullTimeEmployee ; numberFullTimeEmployees++) {
            
//            System.out.println("numberFullTimeEmployees: " + numberFullTimeEmployees);
            
            InitialSchedules.ModelFullTimeEmployeesInitialSchedules myModelFullTimeEmployeesInitialSchedules
                    = new InitialSchedules.ModelFullTimeEmployeesInitialSchedules(myScheduleParameters, numberFullTimeEmployees);

            SolverInitialSchedules mySolverFullTimeEmployeesInitialSchedules = new SolverInitialSchedules(myModelFullTimeEmployeesInitialSchedules);
            List<Solution> myInitialFullTimeEmployeesSchedulesSolutions = mySolverFullTimeEmployeesInitialSchedules.findAllSolution();

            // Création de l'objet banque d'horaire, qui est une liste d'objets de type InitialSchedulesArray, qui contiennent chacun deux tableaux.
            for (Solution CurrentFullTimeEmployeesSolution : myInitialFullTimeEmployeesSchedulesSolutions) {
                FullTimeEmployeeScheduleArray fullTimeSchedulesArray = new FullTimeEmployeeScheduleArray(myModelFullTimeEmployeesInitialSchedules, CurrentFullTimeEmployeesSolution);

                InitialSchedules.ParametersSchedules myPartTimeScheduleParameters = new InitialSchedules.ParametersSchedules();
                myPartTimeScheduleParameters.adjustRequiredWorkforce(fullTimeSchedulesArray);

                InitialSchedules.ModelPartTimeEmployeesInitialSchedules myModelPartTimeEmployeesInitialSchedules
                    = new InitialSchedules.ModelPartTimeEmployeesInitialSchedules(myPartTimeScheduleParameters);

                SolverInitialSchedules mySolverPartTimeEmployeesInitialSchedules = new SolverInitialSchedules(myModelPartTimeEmployeesInitialSchedules);
                List<Solution> myInitialPartTimeEmployeesSchedulesSolutions = mySolverPartTimeEmployeesInitialSchedules.findAllSolution();
//                mySolverPartTimeEmployeesInitialSchedules.printStatistics();

                for (Solution CurrentPartTimeEmployeesSolution : myInitialPartTimeEmployeesSchedulesSolutions) {
                    PartTimeEmployeeScheduleArray partTimeSchedulesArray = new PartTimeEmployeeScheduleArray(myModelPartTimeEmployeesInitialSchedules, CurrentPartTimeEmployeesSolution);
                    SchedulesArray initialSchedulesArray = new SchedulesArray(fullTimeSchedulesArray, partTimeSchedulesArray, myScheduleParameters, simulationNumber);
                    initialSchedulesArrayList.add(initialSchedulesArray);
                    simulationNumber++;
                }
                
            }
            
        }
        
        // Pour afficher les horaires initiales
//        SchedulesWindows.ShowSchedulesFrame initialSchedulesViewer = new SchedulesWindows.ShowSchedulesFrame(initialSchedulesArrayList, "Initial Schedules"); // to be removed

        return initialSchedulesArrayList;

    }

    private static List<BankModifiedSchedules> GenerateModifiedSchedules(List<SchedulesArray> initialSchedulesArrayList, ParametersSchedules myScheduleParameters) {
        
        RandomEngine randAbsence = new MersenneTwister(123456789);
        
        // generation de nouveaux vecteurs de demande de travail
        RandomEngine randVariableDemand = new MersenneTwister(987654321);
        int[][] matrixVariableDemandVector = new int[ParametersVariableDemandSimulator.numberVariableDemandSimulations][myScheduleParameters.getWorkPeriodsPerSchedule()];
        for (int varDemSimNum = 0 ; varDemSimNum < ParametersVariableDemandSimulator.numberVariableDemandSimulations ; varDemSimNum++) {   
            matrixVariableDemandVector[varDemSimNum] = new ModifiedSchedules.VariableDemandVector(myScheduleParameters.getWorkPeriodsPerSchedule(), randVariableDemand).getVariableDemandVector();
        }
        
        List<BankModifiedSchedules> listBankModifiedSchedules = new ArrayList<>();
        List<SchedulesArray> listModifiedSchedules = new ArrayList<>(); // to be removed
        for (SchedulesArray initialSchedule : initialSchedulesArrayList) {
            BankModifiedSchedules bankModificationSimulations = new BankModifiedSchedules(initialSchedule);
            for (int varDemSimNum = 0 ; varDemSimNum < ParametersVariableDemandSimulator.numberVariableDemandSimulations ; varDemSimNum++) {   
                int[] variableDemandVector = matrixVariableDemandVector[varDemSimNum];
                ParametersSchedules myModifiedScheduleParameters = new InitialSchedules.ParametersSchedules(myScheduleParameters);
                myModifiedScheduleParameters.modifyRequiredWorkForce(variableDemandVector);
                for (int absSimNum = 0 ; absSimNum < ParametersAbsencesSimulator.numberAbsenceSimulations ; absSimNum++) {
                    ModifiedSchedulesArray modifiedSchedule = new ModifiedSchedulesArray(initialSchedule, myModifiedScheduleParameters, randAbsence);
                    listModifiedSchedules.add(modifiedSchedule); // to be removed
                    bankModificationSimulations.addModifiedSchedules(modifiedSchedule);
                }
            }
            listBankModifiedSchedules.add(bankModificationSimulations);
        }
        
        // Pour afficher les horaires modifies
//        SchedulesWindows.ShowSchedulesFrame modifiedSchedulesViewer = new SchedulesWindows.ShowSchedulesFrame(listModifiedSchedules, "Modified Schedules"); // to be removed
        
        return listBankModifiedSchedules;
    }
    
    private static List<BankRecoveredSchedules> GenerateOptimalRecoveredSchedules(List<BankModifiedSchedules> listBankModifiedSchedules) {
        
        BankRecoveredSchedulesComparator meanCostComparator = new BankRecoveredSchedulesComparator();
        PriorityQueue< BankRecoveredSchedules > bankRecoveredSchedulesOrderedByMeanCost = new PriorityQueue<>(meanCostComparator);
        List<SchedulesArray> listRecoveredSchedules = new ArrayList<>();  // to be removed
        for (BankModifiedSchedules bankModifiedSchedule : listBankModifiedSchedules) {
            SchedulesArray initialSchedulesArray = bankModifiedSchedule.initialSchedulesArray;
            BankRecoveredSchedules bankRecoveredSimulations = new BankRecoveredSchedules(initialSchedulesArray, bankModifiedSchedule.bankModifiedSimulation);
            int totalCostRecoveredSimulations = 0;
            for (ModifiedSchedulesArray modifiedSchedule : bankModifiedSchedule.bankModifiedSimulation) {
                RecoveredSchedulesArray recoveredSchedule = new RecoveredSchedulesArray(modifiedSchedule);
                listRecoveredSchedules.add(recoveredSchedule); // to be removed
                bankRecoveredSimulations.addRecoveredSchedules(recoveredSchedule);
                if (recoveredSchedule.totalScheduleCost != Integer.MAX_VALUE){
                    totalCostRecoveredSimulations += recoveredSchedule.totalScheduleCost;
                } else {
                    bankRecoveredSimulations.numberUnResolvedSimulations += 1;
                }
                
            }
            bankRecoveredSimulations.meanCostRecoveredSimulations = totalCostRecoveredSimulations / bankRecoveredSimulations.bankRecoveredSimulation.size();
            System.out.println(bankRecoveredSimulations.meanCostRecoveredSimulations);
            bankRecoveredSchedulesOrderedByMeanCost.add(bankRecoveredSimulations);
        }
        // Pour afficher les horaires de recouvrement
//        SchedulesWindows.ShowSchedulesFrame recoveredSchedulesViewer = new SchedulesWindows.ShowSchedulesFrame(listRecoveredSchedules, "Recovered Schedules"); // to be removed
        
        List<BankRecoveredSchedules> listbankRecoveredSchedulesOrderedByMeanCost = new ArrayList<>();
        while(!bankRecoveredSchedulesOrderedByMeanCost.isEmpty()){
            BankRecoveredSchedules bankRecoveredSchedules = bankRecoveredSchedulesOrderedByMeanCost.poll();
            listbankRecoveredSchedulesOrderedByMeanCost.add(bankRecoveredSchedules);
        }
        
        return listbankRecoveredSchedulesOrderedByMeanCost;
    }
}
